/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <dlfcn.h>

#include <string>
#include <fstream>
#include <vector>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "utils.h"
#include "log.h"
#include "launcher.h"
#include "dotnet_launcher.h"
#include "plugin_manager.h"
#include "path_manager.h"
#include "log_manager.h"

#define PLUGIN_PATH "/usr/share/dotnet.tizen/lib/libdotnet_plugin.so"

namespace tizen {
namespace runtime {
namespace dotnetcore {

#if defined (__aarch64__)
#define ARCHITECTURE_IDENTIFIER "arm64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-arm64", "linux", "unix-arm64", "unix", "any", "base"};

#elif defined (__arm__)
#define ARCHITECTURE_IDENTIFIER "arm"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"tizen.4.0.0-armel", "tizen.4.0.0", "tizen-armel", "tizen", "linux-armel", "linux", "unix-armel", "unix", "any", "base"};

#elif defined (__x86_64__)
#define ARCHITECTURE_IDENTIFIER "x64"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-x64", "linux", "unix-x64", "unix", "any", "base"};

#elif defined (__i386__)
#define ARCHITECTURE_IDENTIFIER "x86"
const static std::vector<std::string> RID_FALLBACK_GRAPH =
	{"linux-x86", "linux", "unix-x86", "unix", "any", "base"};

#else
#error "Unknown target"
#endif

static std::string getExtraNativeLibDirs(const std::string& appRoot)
{
	std::string candidate;
	for (unsigned int i = 0; i < RID_FALLBACK_GRAPH.size(); i++) {
		if(!candidate.empty()) {
			candidate += ":";
		}
		candidate += concatPath(appRoot, "bin/runtimes/" + RID_FALLBACK_GRAPH[i] + "/native");
	}

	candidate = candidate + ":" + concatPath(appRoot, "lib/" ARCHITECTURE_IDENTIFIER);
	if (!strncmp(ARCHITECTURE_IDENTIFIER, "arm64", 5)) {
		candidate = candidate + ":" + concatPath(appRoot, "lib/aarch64");
	}

	return candidate;
}

CoreRuntime::CoreRuntime(const char* mode) :
	initializeClr(nullptr),
	executeAssembly(nullptr),
	shutdown(nullptr),
	createDelegate(nullptr),
	__coreclrLib(nullptr),
	__hostHandle(nullptr),
	__domainId(-1),
	fd(0),
	__mode(mode),
	__initialized(false)
{
	_DBG("Constructor called!!");
}

CoreRuntime::~CoreRuntime()
{
	dispose();
}

int CoreRuntime::initialize(bool standalone)
{
#define __XSTR(x) #x
#define __STR(x) __XSTR(x)

#ifdef NATIVE_LIB_DIR
	__nativeLibDirectory = __STR(NATIVE_LIB_DIR);
#endif

#undef __STR
#undef __XSTR

#ifdef __arm__
	// libunwind library is used to unwind stack frame, but libunwind for ARM
	// does not support ARM vfpv3/NEON registers in DWARF format correctly.
	// Therefore let's disable stack unwinding using DWARF information
	// See https://github.com/dotnet/coreclr/issues/6698
	//
	// libunwind use following methods to unwind stack frame.
	// UNW_ARM_METHOD_ALL          0xFF
	// UNW_ARM_METHOD_DWARF        0x01
	// UNW_ARM_METHOD_FRAME        0x02
	// UNW_ARM_METHOD_EXIDX        0x04
	putenv(const_cast<char *>("UNW_ARM_UNWIND_METHOD=6"));
#endif // __arm__

	if (initializePluginManager(__mode) < 0) {
		_ERR("Failed to initialize PluginManager");
		return -1;
	}

	if (initializePathManager(std::string(), std::string(), std::string()) < 0) {
		_ERR("Failed to initialize PathManager");
		return -1;
	}

	if (initializeLogManager() < 0) {
		_ERR("Failed to initnialize LogManager");
		return -1;
	}

	if (redirectFD() < 0) {
		_ERR("Failed to redirect FD");
		return -1;
	}

	if (runLoggingThread() < 0) {
		_ERR("Failed to create and run logging thread to redicrect log");
		return -1;
	}

	std::string libCoreclr(concatPath(getRuntimeDir(), "libcoreclr.so"));

	__coreclrLib = dlopen(libCoreclr.c_str(), RTLD_NOW | RTLD_LOCAL);
	if (__coreclrLib == nullptr) {
		char *err = dlerror();
		_ERR("dlopen failed to open libcoreclr.so with error %s", err);
		return -1;
	}

#define CORELIB_RETURN_IF_NOSYM(type, variable, name) \
	do { \
		variable = (type)dlsym(__coreclrLib, name); \
		if (variable == nullptr) { \
			_ERR(name " is not found in the libcoreclr.so"); \
			return -1; \
		} \
	} while (0)

	CORELIB_RETURN_IF_NOSYM(coreclr_initialize_ptr, initializeClr, "coreclr_initialize");
	CORELIB_RETURN_IF_NOSYM(coreclr_execute_assembly_ptr, executeAssembly, "coreclr_execute_assembly");
	CORELIB_RETURN_IF_NOSYM(coreclr_shutdown_ptr, shutdown, "coreclr_shutdown");
	CORELIB_RETURN_IF_NOSYM(coreclr_create_delegate_ptr, createDelegate, "coreclr_create_delegate");

#undef CORELIB_RETURN_IF_NOSYM

	_DBG("libcoreclr dlopen and dlsym success");

	if (!standalone)
		pluginPreload();

	fd = open("/proc/self", O_DIRECTORY);
	std::string appRoot = std::string("/proc/self/fd/") + std::to_string(fd);
	std::string appBin = concatPath(appRoot, "bin");
	std::string appLib = concatPath(appRoot, "lib");
	std::string probePath = appBin + ":" + appLib;
	std::string tpa = getTPA();
	std::string nativeLibPath = getExtraNativeLibDirs(appRoot) + ":" + appBin + ":" + appLib + ":" + __nativeLibDirectory;
	std::string appName = std::string("dotnet-launcher-") + std::to_string(getpid());

	if (!initializeCoreClr(appName.c_str(), probePath.c_str(), nativeLibPath.c_str(), tpa.c_str())) {
		_ERR("Failed to initialize coreclr");
		return -1;
	}

	__initialized = true;

	return 0;
}

bool CoreRuntime::initializeCoreClr(const char* appId,
									 const char* assemblyProbePaths,
									 const char* pinvokeProbePaths,
									 const char* tpaList)
{
	const char *propertyKeys[] = {
		"TRUSTED_PLATFORM_ASSEMBLIES",
		"APP_PATHS",
		"APP_NI_PATHS",
		"NATIVE_DLL_SEARCH_DIRECTORIES",
		"AppDomainCompatSwitch"
	};

	const char *propertyValues[] = {
		tpaList,
		assemblyProbePaths,
		assemblyProbePaths,
		pinvokeProbePaths,
		"UseLatestBehaviorWhenTFMNotSpecified"
	};

	std::string selfPath = readSelfPath();

	int st = initializeClr(selfPath.c_str(),
							appId,
							sizeof(propertyKeys) / sizeof(propertyKeys[0]),
							propertyKeys,
							propertyValues,
							&__hostHandle,
							&__domainId);

	if (st < 0) {
		_ERR("initialize core clr fail! (0x%08x)", st);
		return false;
	}

	pluginSetCoreclrInfo(__hostHandle, __domainId, createDelegate);

	_DBG("Initialize core clr success");
	return true;
}

void CoreRuntime::dispose()
{
	if (__hostHandle != nullptr) {
		int st = shutdown(__hostHandle, __domainId);
		if (st < 0)
			_ERR("shutdown core clr fail! (0x%08x)", st);
		__hostHandle = nullptr;
	}

	if (__coreclrLib != nullptr) {
		if (dlclose(__coreclrLib) != 0) {
			_ERR("libcoreclr.so close failed");
		}

		__coreclrLib = nullptr;
	}

	finalizePluginManager();
	finalizePathManager();

	_DBG("Dotnet runtime disposed");
}

int CoreRuntime::launch(const char* appId, const char* root, const char* path, int argc, char* argv[])
{
	if (!__initialized) {
		_ERR("Runtime is not initialized");
		return -1;
	}

	if (path == nullptr) {
		_ERR("executable path is null");
		return -1;
	}

	if (!isFileExist(path)) {
		_ERR("File not exist : %s", path);
		return -1;
	}

	// launchpad override stdout and stderr to journalctl before launch application.
	// we have to re-override that to input pipe for logging thread.
	if (redirectFD() < 0) {
		_ERR("Failed to redirect FD");
		return -1;
	}

	pluginSetAppInfo(appId, path);

	int fd2 = open(root, O_DIRECTORY);
	dup3(fd2, fd, O_CLOEXEC);
	if (fd2 >= 0)
		close(fd2);

	pluginBeforeExecute();

	unsigned int ret = 0;
	int st = executeAssembly(__hostHandle, __domainId, argc, (const char**)argv, path, &ret);
	if (st < 0)
		_ERR("Failed to Execute Assembly %s (0x%08x)", path, st);
	return ret;
}

}  // namespace dotnetcore
}  // namespace runtime
}  // namespace tizen
