/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "launcher.h"
#include "log.h"

#include <launchpad.h>
#include <aul.h>

#include <Ecore.h>
#include <Elementary.h>
#include <bundle_internal.h>

#include <map>
#include <vector>
#include <functional>

#include <unistd.h>
#include <dlfcn.h>


namespace tizen {
namespace runtime {

struct FdHandler {
	Ecore_Fd_Handler *handler;
	loader_receiver_cb receiver;
};

static int __argc;
static char **__argv;
static Evas_Object *__win;

class LaunchpadAdapterImpl : public LaunchpadAdapter
{
	public:
		LaunchpadAdapterImpl() :
			callbacks(),
			adapter(),
			__isLaunched(false)
		{ }
		int loaderMain(int argc, char* argv[]) override;

		std::map<int, FdHandler> handlers;

	private:
		AppInfo appInfo;
		loader_lifecycle_callback_s callbacks;
		loader_adapter_s adapter;
		bool __isLaunched;
		std::string __launchPath;
};

LaunchpadAdapterImpl LaunchpadImpl;
LaunchpadAdapter& Launchpad = LaunchpadImpl;

#define WITH_SELF(data) \
	LaunchpadAdapterImpl* self = static_cast<LaunchpadAdapterImpl*>(data); \
	if (self == nullptr) \
		_ERR("No LaunchpadImplData"); \
	else

static Eina_Bool fdHandler(void *data, Ecore_Fd_Handler* handler)
{
	WITH_SELF(data) {
		int fd = ecore_main_fd_handler_fd_get(handler);
		if (fd == -1) {
			_ERR("Failed to get the Ecore FD");
			exit(-1);
		}

		if (ecore_main_fd_handler_active_get(handler, ECORE_FD_READ)) {
			if (self->handlers.find(fd) != self->handlers.end())
				self->handlers[fd].receiver(fd);
		} else if (ecore_main_fd_handler_active_get(handler, ECORE_FD_ERROR)) {
			_ERR("Ecore FD Handler Have Error");
			close(fd);
			exit(-1);
		}
	}

	return ECORE_CALLBACK_CANCEL;
}

static void fdAdd(void *data, int fd, loader_receiver_cb receiver)
{
	Ecore_Fd_Handler* handler = ecore_main_fd_handler_add(fd,
			static_cast<Ecore_Fd_Handler_Flags>(ECORE_FD_READ | ECORE_FD_ERROR),
			fdHandler, data, nullptr, nullptr);
	if (handler == nullptr) {
		_ERR("Failed to add a FD handler to ecore main loop");
		close(fd);
		exit(-1);
	} WITH_SELF(data) {
		self->handlers[fd] = {handler, receiver};
	}
}

static void fdRemove(void *data, int fd)
{
	WITH_SELF(data) {
		if (self->handlers.find(fd) != self->handlers.end()) {
			Ecore_Fd_Handler* handler = self->handlers[fd].handler;
			ecore_main_fd_handler_del(handler);
			self->handlers.erase(fd);
		}
	}
}

static void preCreateWindow(bundle *extra, int type, void *userData)
{
	int elmInitCnt = 0;

	// Precreate window
	elmInitCnt = elm_init(__argc, __argv);
	_DBG("[candidate] elm init, returned: %d", elmInitCnt);

	elm_config_accel_preference_set("hw");

	__win = elm_win_add(NULL, "package_name", ELM_WIN_BASIC);
	if (__win == NULL) {
		_DBG("[candidate] elm_win_add() failed");
		return;
	}

	elm_win_precreated_object_set(__win);
}

int LaunchpadAdapterImpl::loaderMain(int argc, char* argv[])
{
	__argc = argc;
	__argv = argv;
	callbacks.create = [](bundle *extra, int type, void *userData) {
		ecore_init();
		preCreateWindow(extra, type, userData);
		WITH_SELF(userData) {
			if (self->onCreate != nullptr)
				self->onCreate();
		}
	};
	callbacks.launch = [](int argc, char** argv, const char* appPath,
						const char* appId, const char* pkgId,
						const char* pkgType, void* userData) -> int {
		WITH_SELF(userData) {
			const char* appRootPath = aul_get_app_root_path();
			if (appRootPath != nullptr) {
				self->appInfo.root = std::string(appRootPath);
			}
			self->appInfo.path = appPath;
			self->appInfo.id = appId;
			self->appInfo.pkg = pkgId;
			self->appInfo.type = pkgType;
			if (self->onLaunch != nullptr)
				self->onLaunch(self->appInfo, argc, argv);
		}

		return 0;
	};
	callbacks.terminate = [](int argc, char **argv, void* userData) -> int {
		WITH_SELF(userData) {
			if (self->onTerminate != nullptr)
				self->onTerminate(self->appInfo, argc, argv);
		}
		return 0;
	};

	adapter.loop_begin = [](void *data) {
		ecore_main_loop_begin();
	};

	adapter.loop_quit = [](void *data) {
		ecore_main_loop_quit();
	};
	adapter.add_fd = fdAdd;
	adapter.remove_fd = fdRemove;

	_DBG("launchpad_loader_main is start");
	int r = launchpad_loader_main(argc, argv, &(this->callbacks), &(this->adapter), this);
	_DBG("launchpad_loader_main is finished with [%d]", r);

	return r;
}

#undef WITH_SELF

}  // namespace runtime
}  // namespace tizen
